/*****************************************************************
 * BK Instruments Polling Device
 *
 *  Copyright (c) 2012
 *   Vanderhaegen Bart
 *
 *
 * This device will instruct the PIC16F628 to read and poll the
 * BK Instrument every 100 milliseconds to detect an over-current.
 *
 *
 *****************************************************************/

// Includes
#include "htc.h";

// Definitions
#define _XTAL_FREQ 19668000
#define PACKET_SIZE 26
#define PACKET_ADDR 0
#define LED_OK RB0
#define LED_ERR	RB5
#define BUZZER RB6

// Variable Declarations
char packetTransmit[PACKET_SIZE];	// transmission buffer
char packetReceived[PACKET_SIZE];	// receive buffer
char receiveCounter=0;				// number of received bytes

// Declaration of the error flags
bit framingErrorOccured=0;
bit invalidFrameLengthOccured=0;
bit invalidCheckSumOccured=0;
bit invalidResponseOccured=0;
bit frameReceived=0;


// checkSum() returns the checksum of the received data
char checkSum(void) {
	int checkSum=0;
	for (char i=0;i<PACKET_SIZE-2;i++) {
		checkSum+=packetReceived[i];
	}
	return checkSum % 256;
}




// is_bit_set checks if a certain bit of a byte is 1 or 0.
bit is_bit_set(char value, char bitindex)
{
    if ((value & (1 << bitindex)) != 0)
		return 1;
	else
		return 0;
}




// processResponse() processes the packets when they are received.
void processResponse() {

	frameReceived=1;

	if (receiveCounter==PACKET_SIZE) {
		if (checkSum()==packetReceived[PACKET_SIZE-1]) {
			// data valid (right length & checksum matches)
			LED_OK=!LED_OK;
			LED_ERR=0;
			
			if (packetReceived[2]==0x26) {
				// Evaluate CC & CV mode
				if(is_bit_set(packetReceived[9],3)==1)
					BUZZER=1;
				else
					BUZZER=0;
			}
			else {
				invalidResponseOccured=1;
			}
		}
		else {
			// Error: invalid checksum
			invalidCheckSumOccured=1;
		}
	}
	else {
		// Error: invalid data length
		invalidFrameLengthOccured=1;
	}
}




static void interrupt isr( void ) {
	if (RCIF) { 
		if (FERR==1) {
			// A framing error occured.
			framingErrorOccured=1;
			FERR=0;
		}

		// check the overrun error bit
		if (OERR==1) {
			CREN=0;
			CREN=1;
		}

		// Receive the data packet
		packetReceived[receiveCounter]=RCREG;
		receiveCounter++;

		// stop receiving data when 26 bytes are received
		if (receiveCounter>=PACKET_SIZE) {
			processResponse();
			receiveCounter=0;
		}
	}
}




// sendRequest() will send the request frame to the BK unit.
void sendRequest(void) {
	TXEN=1;
	for (char i=0;i<PACKET_SIZE;i++) {
		while(TXIF==0);
		TXREG=packetTransmit[i];
	}
	receiveCounter=0;
}




// initUART() will initialise the UART of the PIC
// The baud rate is set to baud 9600, the X-TAL is 19.668 MHz.
void initUART() {
	// Configure the baud rate
	SPBRG = 127; 	// baud rate generator register
	BRGH = 1;		// Enable high baud-rate

	// Enable asynchronous communication
	SYNC = 0;		// Asynchronous mode
	SPEN = 1;		// Enable serial port
	TRISB1 = 1;		// Required by UART (datasheet)
	TRISB2 = 1;		// Required by UART (datasheet)

	// Enable 8-bit data words
	TX9 = 0;
	RX9 = 0;

	// Enable Interrupts
	RCIE = 1;  		// enable USART interrupt
	CREN = 1;  		// Continuous recieve enable bit : ENABLE
	PEIE = 1;  		// enable unmasked periheral interrupts
	GIE = 1;   		// global interrupt enable
}



// executeStartupTest() runs a small test when starting the PIC.
void executeStartupTest() {
	// Execute startup test
	for (char i=0;i<4;i++) {
		__delay_ms(100);
		LED_OK=1;
		LED_ERR=1;
		BUZZER=1;
		__delay_ms(100);
		LED_OK=0;
		LED_ERR=0;
		BUZZER=0;
	}
}

// flashError() flashes the LEDs for a number of times.
void flashError(char numberOfFlashes) {
	LED_ERR=0;

	for (char i=0;i<numberOfFlashes;i++) {
		LED_ERR=!LED_ERR;
		__delay_ms(500);
	}

	__delay_ms(1000);
}



// pollErrorFags() will trigger the error LEDs
// Note: in the current program only one error can occur at once.
void pollErrorFlags() {

	if (frameReceived==0)
		flashError(2);	// no response

	if (framingErrorOccured==1)
		flashError(4);	// framing errors (RS-232 error)

	if (invalidFrameLengthOccured==1)
		flashError(6);	// invalid frame received

	if (invalidCheckSumOccured==1)
		flashError(8);	// invalid checksum calculated

	if (invalidResponseOccured==1)
		flashError(10);	// invalid response
}



// initTransmissionPacket() will prepare the packetTransmit array.
void initTransmissionPacket() {
	for (char i=0;i<PACKET_SIZE;i++)
		packetTransmit[i]=0;

	packetTransmit[0]=0xAA;
	packetTransmit[2]=0x26;
	packetTransmit[25]=0xD0;
}


// MAIN PROGRAM
void main() {

	// Initialisation of the I/O
	TRISB = 0x00;
	PORTB = 0x00;

	// Initialisation of the UART
	initUART();

	// initialisation of the data packet
	initTransmissionPacket();

	// Execute startup test
	executeStartupTest();


	// Main program loop
	// This loop is cycled until the end of days
	while(1) {
		// Reset the LEDs
		FERR=0;
		LED_OK=0;
		LED_ERR=0;

		// Reset the error messages
		framingErrorOccured=0;
		invalidFrameLengthOccured=0;
		invalidCheckSumOccured=0;
		invalidResponseOccured=0;
		frameReceived=0;

		// Send a request to the BK unit
		sendRequest();
		
		// Add a delay before executing the next cycle. 
		// Data is received during this delay (interrupts)
		__delay_ms(100);

		// poll the error flags
		pollErrorFlags();	
	}
}


